from jennifer.agent import jennifer_agent
from jennifer.api.proxy import Proxy


def safe(l, idx, default=None):
    try:
        return l[idx]
    except IndexError:
        return default


class CursorProxy(Proxy):
    def __init__(self, obj, host, port, paramstyle, conn):
        Proxy.__init__(self, obj)
        self.set('host', host)
        self.set('port', port)
        self.set('paramstyle', paramstyle)
        self.set('conn', conn)

    def execute(self, *args, **kwargs):
        agent = jennifer_agent()
        transaction = jennifer_agent().current_transaction()
        operation = safe(args, 0) or kwargs.get('operation')
        parameters = safe(args, 1) or kwargs.get('parameters')
        agent.recorder.db_recorder.active(self.conn)
        if transaction is not None and operation is not None:
            if parameters is not None and \
                (self.paramstyle == 'format' or self.paramstyle == 'pyformat'):
                operation = operation % (parameters)
            transaction.profiler.db_execute(self.host, self.port, operation)
        result = None
        try:
            result = self._origin.execute(*args, **kwargs)
        except Exception as e:
            if transaction is not None and operation is not None:
                transaction.profiler.sql_error(e)
                transaction.profiler.end()
            agent.recorder.db_recorder.deactive(self.conn)
            raise e

        if transaction is not None and operation is not None:
            transaction.profiler.end()
        agent.recorder.db_recorder.deactive(self.conn)

        return result

    def record_fetch(self, fetch, size, pass_size=False):
        transaction = jennifer_agent().current_transaction()
        agent = jennifer_agent()
        args = []
        if pass_size:
            args = [size]
        if transaction is None:
            try:
                agent.recorder.db_recorder.active(self.conn)
                ret = fetch(*args)
            except Exception as e:
                agent.recorder.db_recorder.deactive(self.conn)
                raise e
            agent.recorder.db_recorder.deactive(self.conn)
            return ret

        if size == 0:  # it mean fetch all
            size = self._result.num_rows()
        transaction.profiler.db_fetch(size)
        err = None
        try:
            agent.recorder.db_recorder.active(self.conn)
            ret = fetch(*args)
        except Exception as e:
            err = e
        transaction.profiler.end()
        agent.recorder.db_recorder.deactive(self.conn)
        if err is not None:
            raise err
        return ret

    def fetchone(self):
        return self.record_fetch(self._origin.fetchone, 1)

    def fetchmany(self, size=None):
        pass_size = True
        if size is None:
            size = self._origin.arraysize
            pass_size = False
        return self.record_fetch(self._origin.fetchmany, size, pass_size)

    def fetchall(self):
        size = self.rowcount
        return self.record_fetch(self._origin.fetchall, size)


class ConnectionProxy(Proxy):
    def __init__(self, obj, host, port, paramstyle):
        Proxy.__init__(self, obj)
        self.set('host', host)
        self.set('port', port)
        self.set('paramstyle', paramstyle)

    def cursor(self, *args, **kwargs):
        cursor = self._origin.cursor(*args, **kwargs)
        return CursorProxy(cursor, self.host, self.port, self.paramstyle, self)

    def close(self, *args, **kwargs):
        jennifer_agent().recorder.db_recorder.remove_connection(self)
        return self._origin.close(*args, **kwargs)


def register_database(module, connection_info):
    agent = jennifer_agent()
    def wrap_connect(connect):
        def handler(*args, **kwargs):
            host, port, database = connection_info(*args, **kwargs)
            transaction = agent.current_transaction()
            if transaction is not None:
                transaction.profiler.db_open(host, port, database)
            try:
                connection = ConnectionProxy(
                    connect(*args, **kwargs), host, port, module.paramstyle)
            except Exception as e:
                if transaction is not None:
                    transaction.profiler.db_connection_error(e)
                    transaction.profiler.end()
                raise e
            if transaction is not None:
                transaction.profiler.end()
            agent.recorder.db_recorder.add_connection(connection)
            return connection
        return handler
    module.connect = wrap_connect(module.connect)



from random import random
from jennifer.agent import jennifer_agent
from .profiler import TransactionProfiler



class Transaction:
    def __init__(self, agent, start_time, environ, wmonid, *args, **kwargs):
        self.init_property(*args, **kwargs)
        self.agent = agent
        self.start_system_time = 0
        self.end_system_time = 0
        self.wmonid = wmonid
        self.thread_id = jennifer_agent().current_thread_id()
        if environ.get('HTTP_X_FORWARDED_FOR') is not None:
            self.client_address = environ.get('HTTP_X_FORWARDED_FOR')
        elif environ.get('HTTP_CLIENT_IP') is not None:
            self.client_address = environ.get('HTTP_CLIENT_IP')
        else:
            self.client_address = environ.get('REMOTE_ADDR', '')
        self.browser_info_hash = self.agent.hash_text(
            environ.get('HTTP_USER_AGENT', ''), 'browser_info')
        path_info = environ.get(
            'PATH_INFO', '').encode('iso-8859-1').decode('utf-8')
        self.request_method = environ.get('REQUEST_METHOD', '')
        self.query_string = environ.get('QUERY_STRING', '')
        self.start_time = start_time
        self.path_info = path_info
        self.service_hash = self.agent.hash_text(path_info)
        self.profiler = TransactionProfiler(self, self.service_hash)

    def init_property(
            self,
            txid=0,
            elapsed=0,
            elapsed_cpu=0,
            end_time=0,
            sql_count=0,
            sql_time=0,
            fetch_count=0,
            fetch_time=0,
            external_call_count=0,
            external_call_time=0,
            client_address=None,
            wmonid=0,
            user_hash=0,
            service_hash=0,
            guid=None,
            browser_info_hash=0,
            error_code=0,
    ):
        self.elapsed = elapsed
        self.elapsed_cpu = elapsed_cpu
        self.end_time = end_time
        self.sql_count = sql_count
        self.sql_time = sql_time
        self.fetch_count = fetch_count
        self.fetch_time = fetch_time
        self.external_call_count = external_call_count
        self.external_call_time = external_call_time
        self.txid = txid
        self.client_address = client_address
        self.wmonid = wmonid
        self.user_hash = user_hash
        self.guid = guid
        self.browser_info_hash = browser_info_hash
        self.error_code = error_code
        self.service_hash = service_hash
        self.incoming_remote_call = None

    def end(self, end_time=None):
        self.profiler.end()  # end root method
        self.end_system_time = self.agent.current_cputime()
        self.end_time = end_time
        if self.end_time is None:
            self.end_time = self.agent.current_time()
        self.agent.end_transaction(self)

    def to_active_service_dict(self, current_time, current_cputime):
        self.elapsed_cpu = int(current_cputime - self.start_system_time)
        self.elapsed = int(current_time - self.start_time)
        data = { key: getattr(self, key) for key in [
            'elapsed',
            'elapsed_cpu',
            'sql_count',
            'fetch_count',
            'txid',
            'client_address',
            'wmonid',
            'service_hash',
            'thread_id',
            'start_time',
        ]}

        data['running_mode'] = self.profiler.running_mode
        data['running_hash'] = self.profiler.running_hash
        data['status_code'] = self.profiler.status
        data['running_time'] = current_time - self.profiler.running_start_time

        return data

    def to_dict(self):
        self.elapsed_cpu = int(self.end_system_time - self.start_system_time)
        self.elapsed = int(self.end_time - self.start_time)
        data = { key: getattr(self, key) for key in [
            'elapsed',
            'elapsed_cpu',
            'end_time',
            'sql_count',
            'sql_time',
            'fetch_count',
            'fetch_time',
            'external_call_count',
            'external_call_time',
            'txid',
            'client_address',
            'wmonid',
            'user_hash',
            'guid',
            'browser_info_hash',
            'error_code',
            'service_hash',
            'start_system_time',
            'end_system_time',
            'thread_id',
            'start_time',
        ]}

        return data



from jennifer.agent import jennifer_agent
__hooking_module__ = 'urllib'


def wrap_urlopen(urlopen):
    agent = jennifer_agent()
    def handler(*args, **kwargs):  # TODO: is safe in Python2?
        from urllib import parse
        from urllib.request import Request
        from http.client import HTTPResponse
        transaction = agent.current_transaction()
        url = kwargs.get('url') or args[0]
        if isinstance(url, Request):
            url = url.full_url
        if transaction is not None:
            o = parse.urlparse(url)
            transaction.profiler.external_call(
                protocol=o.scheme,
                host=o.hostname,
                port=o.port or 80,
                url=url,
                caller='urllib.request.urlopen',
            )
        ret = urlopen(*args, **kwargs)
        if isinstance(ret, HTTPResponse):
            v = ret.version
            version = 'HTTP/1.1'
            if v is 10:
                version  = 'HTTP/1.0'
            transaction.profiler.end(
                message='uurllib.request.urlopen(url=%s,response=%s,%s)' % (
                    url, version, ret.status)
            )
        else:
            transaction.profiler.end()
        return ret
    return handler


def hook(urllib):
    import urllib.request
    urllib.request.urlopen = wrap_urlopen(urllib.request.urlopen)



from .profile_data import ProfileDataType


class Message(ProfileDataType):
    def __init__(
            self,
            message='',
    ):
        ProfileDataType.__init__(self)
        self.message = message

    def get_type(self):
        return ProfileDataType.TYPE_MESSAGE



from .method import Method


class Root(Method):
    def __init__(
            self,
            name_hash=0,
            error_hash=0,
    ):
        Method.__init__(self, name_hash, error_hash)
        self.parent_index = -1
        self.index = 0



from .gc import GCRecorder
from .db import DBConnectionRecorder


class Recorder(object):
    def __init__(self):
        self.gc_recorder = GCRecorder()
        self.db_recorder = DBConnectionRecorder()

    def record_self(self):
        gc_time, gc_count, gc_gen_time = self.gc_recorder.record()
        db_connection = self.db_recorder.record()

        return {
            'gc_time': gc_time,
            'gc_count': gc_count,
            'gc_gen_time': gc_gen_time,
            'db_connection': db_connection,
        }



from ... import remote_call
from ..profile_data import ProfileDataType


class Query(ProfileDataType):
    TYPE_DB_UNKNOWN = remote_call.TYPE_UNKNOWN_SQL_DATABASE
    TYPE_DB_ORACLE = remote_call.TYPE_ORACLE
    TYPE_DB_DB2 = remote_call.TYPE_DB2
    TYPE_DB_MYSQL = remote_call.TYPE_MYSQL
    TYPE_DB_POSTGRESQL = remote_call.TYPE_POSTGRESQL
    TYPE_DB_CUBRID = remote_call.TYPE_CUBRID
    TYPE_DB_MSSQL = remote_call.TYPE_MSSQL
    TYPE_DB_SOAP = remote_call.TYPE_SOAP

    TYPE_SQL_PSTMT_QUERY = 5
    TYPE_SQL_QUERY = 11
    TYPE_SQL_EXECUTE = 12

    def __init__(
            self,
            host='',
            port='',
            db=TYPE_DB_UNKNOWN,
            method=TYPE_SQL_QUERY,
            query_hash=0,
            error_hash=0,
            bind_params=[],
            inline_params=[],
            query='',
    ):
        ProfileDataType.__init__(self)
        self.error_hash = error_hash
        self.db = db
        self.host = host
        self.port = port
        self.method = method
        self.query_hash = query_hash
        self.bind_params = bind_params
        self.inline_params = inline_params
        self.query = query

    def get_type(self):
        return ProfileDataType.TYPE_SQL_EXEC



from jennifer.agent import jennifer_agent
__hooking_module__ = 'requests'


def wrap_send(origin):
    agent = jennifer_agent()
    def handler(self, request, **kwargs):
        from urllib import parse
        transaction = agent.current_transaction()
        if transaction is not None:
            url = request.url
            o = parse.urlparse(url)
            transaction.profiler.external_call(
                protocol=o.scheme,
                host=o.hostname,
                port=o.port or 80,
                url=url,
                caller='requests.Session.send',
            )
        ret = origin(self, request, **kwargs)
        if transaction is not None:
            message = None
            if ret is not None:
                message='rrequests.Session.send(url=%s,response=%s)' % (
                    url, ret.status_code)
            transaction.profiler.end(message=message)
        return ret
    return handler


def hook(requests):
    requests.Session.send = wrap_send(requests.Session.send)



from jennifer.agent import jennifer_agent
__hooking_module__ = 'MySQLdb'


def safe(l, idx, default=None):
    try:
        return l[idx]
    except IndexError:
        return default

def connection_info(*args, **kwargs):
    host = safe(args, 0) or kwargs.get('host')
    port = safe(args, 6) or kwargs.get('port') or 3306
    database = safe(args, 4) or \
        kwargs.get('database') or kwargs.get('db')

    return host, port, database

def hook(MySQLdb):
    from jennifer.wrap import db_api
    db_api.register_database(MySQLdb, connection_info)



import os

cached_agent = None


def jennifer_agent():
    global cached_agent
    if cached_agent is None:
        from .agent import Agent
        cached_agent = Agent()
        Agent.instance = cached_agent
    return cached_agent



from .profile_data import ProfileDataType


class Socket(ProfileDataType):
    TYPE_UNKNOWN = 0
    TYPE_ISRTREAM = 3
    TYPE_OSRTREAM = 3
    TYPE_IOSTREAM = 7

    def __init__(
            self,
            host,
            port,
            local_port,
    ):
        ProfileDataType.__init__(self)
        self.host = host
        self.port = port
        self.local_port = local_port
        self.mode = Socket.TYPE_IOSTREAM

    def get_type(self):
        return ProfileDataType.TYPE_SOCKET
